"""Raspberry Pi Face Recognition Coffee Maker
Capture and recognize script for registering and training a new user to system
Copyright 2016 Andrew Palmer (ajp294) 

"""

import glob
import os
import sys
import select
import io
import time

import cv2
import numpy as np

import config
import face
import user
import train


# Prefix for positive training image filenames.
POSITIVE_FILE_PREFIX = 'positive_'


def is_letter_input(letter):
   # Utility function to check if a specific character is available on stdin.
   # Comparison is case insensitive.
   if select.select([sys.stdin,],[],[],0.0)[0]:
      input_char = sys.stdin.read(1)
      return input_char.lower() == letter.lower()
   else:
      return False

def main():
  
   # Initialize PiCamera
   camera = config.get_camera()
   
   ###############################
   # Adding new user to database #
   ###############################

   # Ask for name of user who's training
   user_name = raw_input('Please enter username, without spaces, for registration to User Database: ')
   # new_user_attributes.append(user_name)
   print("you entered " + user_name + " as your username")

   # Create new user object with inputted attributes
   new_user = user.User(user_name)

   # Add new user to users file
   # Create file if does not already exist
   if not os.path.exists(config.USERS_FILE):
      new_user.addToFile(config.USERS_FILE, 'w')
   else:
      new_user.addToFile(config.USERS_FILE, 'a')


   # Create the directory for specific user's positive training images if it doesn't exist
   if not os.path.exists(config.POSITIVE_DIR + '_' + new_user.name):
   	os.makedirs(config.POSITIVE_DIR + '_' + new_user.name)

   # Find the largest ID of existing positive images for specific user
   # Start new images after this ID value.
   files = sorted(glob.glob(os.path.join(config.POSITIVE_DIR + '_' + new_user.name, 
   	POSITIVE_FILE_PREFIX + '[0-9][0-9][0-9].pgm')))
   count = 0

   if len(files) > 0:
   	# Grab the count from the last filename.
   	count = int(files[-1][-7:-4])+1


   print 'Capturing positive training images.'
   print 'Type c (and press enter) to capture an image.'
   print 'Enter the letter f to finish image capture and train model.'
   print 'Press Ctrl-C to quit.'
   # TODO FINDME: eventually want to have camera capture say 20 images in a row using the delay func

   while True:
      # camera.start_preview()

      ########################################
      # Capture training images for new user #
      ########################################

      # Check if 'c' or 'f' was received, then act accordingly
      command_input = raw_input("Enter 'c' to capture or 'f' to finish and train: ")

      # if is_letter_input('c'):
      if command_input == 'c':

         print 'Capturing image...'
         image = camera.read()

         # Convert image to grayscale.
         image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

         # FINDME: TODO Add timer here to get detection speed
         # start = time.time()
         start = time.clock()
         # Get coordinates of single face in captured image.
         result = face.detect_single(image)
         # detection_time = time.time() - start
         detection_time = time.clock() - start
         print("Detected Face in: " + str(detection_time) + " seconds")

         if result is None:
         	print 'Could not detect valid face! Please try again'
         	continue
         x, y, w, h = result

         # Crop image as close as possible to desired face aspect ratio.
         # Might be smaller if face is near edge of image.
         crop = face.crop(image, x, y, w, h)

         # Save image to file specific user file
         filename = os.path.join(config.POSITIVE_DIR + '_' + new_user.name, POSITIVE_FILE_PREFIX + '%03d.pgm' % count)
         cv2.imwrite(filename, crop)

         print 'Found face and wrote training image', filename
         count += 1

      # elif (is_letter_input('f')):
      elif command_input == 'f':
         
         ################################################
         # Finished image capture; Start training model #
         ################################################

         # Place positive and negative images into arrays
         print("Reading Training Images...")
         faces = []
         labels = []
         pos_count = 0
         neg_count = 0

         # Read all user specific positive images
         for filename in train.Training().walk_files(config.POSITIVE_DIR + '_' + new_user.name, '*.pgm'):
            faces.append(train.Training().prepare_image(filename))
            labels.append(config.POSITIVE_LABEL)
            pos_count += 1

         # Read all user specific negative images
         for filename in train.Training().walk_files(config.NEGATIVE_DIR, '*.pgm'):
            faces.append(train.Training().prepare_image(filename))
            labels.append(config.NEGATIVE_LABEL)
            neg_count += 1

         print("Read " + str(pos_count) + ' positive images and ' + str(neg_count) + ' negative images')

         # Train Model
         # FINDME: TODO Add timer here to get training speed
         print("Training model...")
         # start = time.time()
         start = time.clock()

         model = cv2.createEigenFaceRecognizer()
         model.train(np.asarray(faces), np.asarray(labels))
         training_duration = time.clock() - start
         print("Trained model in: " + str(training_duration) + " seconds")

         # Save model results
         model.save(config.TRAINING_FILE + '_' + new_user.name + '.xml')
         print("Training data saved to", config.TRAINING_FILE + '_' + new_user.name + '.xml')
         
         # Break execution when training complete
         break



if __name__ == '__main__': main()
